@@ -22,13 +22,13 @@
 """
 Module's constants for the modes of operation supported with CAST:
 
-:var MODE_ECB: Electronic Code Book (ECB)
-:var MODE_CBC: Cipher-Block Chaining (CBC)
-:var MODE_CFB: Cipher FeedBack (CFB)
-:var MODE_OFB: Output FeedBack (OFB)
-:var MODE_CTR: CounTer Mode (CTR)
-:var MODE_OPENPGP:  OpenPGP Mode
-:var MODE_EAX: EAX Mode
+:var MODE_ECB: :ref:`Electronic Code Book (ECB) <ecb_mode>`
+:var MODE_CBC: :ref:`Cipher-Block Chaining (CBC) <cbc_mode>`
+:var MODE_CFB: :ref:`Cipher FeedBack (CFB) <cfb_mode>`
+:var MODE_OFB: :ref:`Output FeedBack (OFB) <ofb_mode>`
+:var MODE_CTR: :ref:`CounTer Mode (CTR) <ctr_mode>`
+:var MODE_OPENPGP:  :ref:`OpenPGP Mode <openpgp_mode>`
+:var MODE_EAX: :ref:`EAX Mode <eax_mode>`
 """
 
 import sys
@@ -37,7 +37,7 @@ from Crypto.Cipher import _create_cipher
 from Crypto.Util.py3compat import byte_string
 from Crypto.Util._raw_api import (load_pycryptodome_raw_lib,
                                   VoidPointer, SmartPointer,
-                                  c_size_t, expect_byte_string)
+                                  c_size_t, c_uint8_ptr)
 
 _raw_cast_lib = load_pycryptodome_raw_lib(
                     "Crypto.Cipher._raw_cast",
@@ -66,8 +66,6 @@ def _create_base_cipher(dict_parameters):
     except KeyError:
         raise TypeError("Missing 'key' parameter")
 
-    expect_byte_string(key)
-
     if len(key) not in key_size:
         raise ValueError("Incorrect CAST key length (%d bytes)" % len(key))
 
@@ -75,7 +73,7 @@ def _create_base_cipher(dict_parameters):
     stop_operation = _raw_cast_lib.CAST_stop_operation
 
     cipher = VoidPointer()
-    result = start_operation(key,
+    result = start_operation(c_uint8_ptr(key),
                              c_size_t(len(key)),
                              cipher.address_of())
     if result:
@@ -91,14 +89,14 @@ def new(key, mode, *args, **kwargs):
     :param key:
         The secret key to use in the symmetric cipher.
         Its length can vary from 5 to 16 bytes.
-    :type key: byte string
+    :type key: bytes, bytearray, memoryview
 
     :param mode:
         The chaining mode to use for encryption or decryption.
     :type mode: One of the supported ``MODE_*`` constants
 
     :Keyword Arguments:
-        *   *iv* (``byte string``) --
+        *   **iv** (*bytes*, *bytearray*, *memoryview*) --
             (Only applicable for ``MODE_CBC``, ``MODE_CFB``, ``MODE_OFB``,
             and ``MODE_OPENPGP`` modes).
 
@@ -114,7 +112,7 @@ def new(key, mode, *args, **kwargs):
             If not provided, a random byte string is generated (you must then
             read its value with the :attr:`iv` attribute).
 
-        *   *nonce* (``byte string``) --
+        *   **nonce** (*bytes*, *bytearray*, *memoryview*) --
             (Only applicable for ``MODE_EAX`` and ``MODE_CTR``).
 
             A value that must never be reused for any other encryption done
@@ -128,17 +126,17 @@ def new(key, mode, *args, **kwargs):
             If not provided for ``MODE_EAX``, a random byte string is generated (you
             can read it back via the ``nonce`` attribute).
 
-        *   *segment_size* (``integer``) --
+        *   **segment_size** (*integer*) --
             (Only ``MODE_CFB``).The number of **bits** the plaintext and ciphertext
             are segmented in. It must be a multiple of 8.
             If not specified, it will be assumed to be 8.
 
-        *   *mac_len* : (``integer``) --
+        *   **mac_len** : (*integer*) --
             (Only ``MODE_EAX``)
             Length of the authentication tag, in bytes.
             It must be no longer than 8 (default).
 
-        *   *initial_value* : (``integer``) --
+        *   **initial_value** : (*integer*) --
             (Only ``MODE_CTR``). The initial value for the counter within
             the counter block. By default it is **0**.
 
@@ -158,4 +156,4 @@ MODE_EAX = 9
 # Size of a data block (in bytes)
 block_size = 8
 # Size of a key (in bytes)
-key_size = xrange(5, 16 + 1)
+key_size = range(5, 16 + 1)
