@@ -73,7 +73,7 @@ EXPORT_SYM int chacha20_init(stream_state **pState,
     if (NULL == key || keySize != KEY_SIZE)
         return ERR_KEY_SIZE;
 
-    if (nonceSize != 8 && nonceSize != 12)
+    if (nonceSize != 8 && nonceSize != 12 && nonceSize != 16)
         return ERR_NONCE_SIZE;
 
     *pState = hs = (stream_state*) calloc(1, sizeof(stream_state));
@@ -84,20 +84,63 @@ EXPORT_SYM int chacha20_init(stream_state **pState,
     hs->h[1] = 0x3320646e;
     hs->h[2] = 0x79622d32;
     hs->h[3] = 0x6b206574;
-    
+
     /** Move 256-bit/32-byte key into h[4..11] **/
     for (i=0; i<32/4; i++) {
         hs->h[4+i] = LOAD_U32_LITTLE(key + 4*i);
     }
-    /** h[12] remains 0 (offset) **/
-    if (nonceSize == 8) {
-        /** h[13] remains 0 (offset) **/
-        hs->h[14] = LOAD_U32_LITTLE(nonce + 0);
-        hs->h[15] = LOAD_U32_LITTLE(nonce + 4);
-    } else {
-        hs->h[13] = LOAD_U32_LITTLE(nonce + 0);
-        hs->h[14] = LOAD_U32_LITTLE(nonce + 4);
-        hs->h[15] = LOAD_U32_LITTLE(nonce + 8);
+
+    switch (nonceSize) {
+    case 8: {
+                /*
+                cccccccc  cccccccc  cccccccc  cccccccc
+                kkkkkkkk  kkkkkkkk  kkkkkkkk  kkkkkkkk
+                kkkkkkkk  kkkkkkkk  kkkkkkkk  kkkkkkkk
+                bbbbbbbb  BBBBBBBB  nnnnnnnn  nnnnnnnn
+
+                c=constant k=key b=blockcount(low) B=blockcount(high) n=nonce
+                */
+
+                /** h[12] remains 0 (offset) **/
+                /** h[13] remains 0 (offset) **/
+                hs->h[14] = LOAD_U32_LITTLE(nonce + 0);
+                hs->h[15] = LOAD_U32_LITTLE(nonce + 4);
+                break;
+                }
+    case 12: {
+                /*
+                cccccccc  cccccccc  cccccccc  cccccccc
+                kkkkkkkk  kkkkkkkk  kkkkkkkk  kkkkkkkk
+                kkkkkkkk  kkkkkkkk  kkkkkkkk  kkkkkkkk
+                bbbbbbbb  nnnnnnnn  nnnnnnnn  nnnnnnnn
+
+                c=constant k=key b=blockcount n=nonce
+                */
+
+                /** h[12] remains 0 (offset) **/
+                hs->h[13] = LOAD_U32_LITTLE(nonce + 0);
+                hs->h[14] = LOAD_U32_LITTLE(nonce + 4);
+                hs->h[15] = LOAD_U32_LITTLE(nonce + 8);
+                break;
+            }
+    case 16: {
+                /*
+                cccccccc  cccccccc  cccccccc  cccccccc
+                kkkkkkkk  kkkkkkkk  kkkkkkkk  kkkkkkkk
+                kkkkkkkk  kkkkkkkk  kkkkkkkk  kkkkkkkk
+                nnnnnnnn  nnnnnnnn  nnnnnnnn  nnnnnnnn
+
+                c=constant k=key n=nonce
+                */
+
+                hs->h[12] = LOAD_U32_LITTLE(nonce + 0);
+                hs->h[13] = LOAD_U32_LITTLE(nonce + 4);
+                hs->h[14] = LOAD_U32_LITTLE(nonce + 8);
+                hs->h[15] = LOAD_U32_LITTLE(nonce + 12);
+                break;
+            }
+    default:
+             return ERR_NONCE_SIZE;
     }
 
     hs->nonceSize = nonceSize;
@@ -114,12 +157,11 @@ EXPORT_SYM int chacha20_destroy(stream_state *state)
     return 0;
 }
 
-static int chacha20_core(stream_state *state)
+static int chacha20_core(stream_state *state, uint32_t h[16])
 {
     unsigned i;
-    uint32_t h[16];
 
-    memcpy(h, state->h, sizeof h);
+    memcpy(h, state->h, sizeof state->h);
 
     for (i=0; i<10; i++) {
         /** Column round **/
@@ -134,27 +176,37 @@ static int chacha20_core(stream_state *state)
         QR(h[3], h[4], h[ 9], h[14]);
     }
 
-    for (i=0; i<16; i++)
-        h[i] += state->h[i];
-
     for (i=0; i<16; i++) {
-        STORE_U32_LITTLE(state->keyStream + 4*i, h[i]);
+        uint32_t sum;
+
+        sum = h[i] + state->h[i];
+        STORE_U32_LITTLE(state->keyStream + 4*i, sum);
     }
 
     state->usedKeyStream = 0;
 
-    if (state->nonceSize == 8) {
-        /** Nonce is 64 bits, counter is two words **/
-        if (++state->h[12] == 0) {
-            if (++state->h[13] == 0) {
-                return ERR_MAX_DATA;
+    switch (state->nonceSize) {
+    case 8: {
+                /** Nonce is 64 bits, counter is two words **/
+                if (++state->h[12] == 0) {
+                    if (++state->h[13] == 0) {
+                        return ERR_MAX_DATA;
+                    }
+                }
+                break;
+            }
+    case 12: {
+                /** Nonce is 96 bits, counter is one word **/
+                if (++state->h[12] == 0) {
+                    return ERR_MAX_DATA;
+                }
+                break;
+            }
+    case 16: {
+                 /** Nonce is 192 bits, there is no counter as this is intended
+                  * to be run once only (HChaCha20) **/
+                 break;
             }
-        }
-    } else {
-        /** Nonce is 96 bits, counter is one word **/
-        if (++state->h[12] == 0) {
-            return ERR_MAX_DATA;
-        }
     }
 
     return 0;
@@ -168,14 +220,18 @@ EXPORT_SYM int chacha20_encrypt(stream_state *state,
     if (NULL == state || NULL == in || NULL == out)
         return ERR_NULL;
 
+    if ((state->nonceSize != 8) && (state->nonceSize != 12))
+        return ERR_NONCE_SIZE;
+
     while (len>0) {
         unsigned keyStreamToUse;
         unsigned i;
+        uint32_t h[16];
 
         if (state->usedKeyStream == sizeof state->keyStream) {
             int result;
 
-            result = chacha20_core(state);
+            result = chacha20_core(state, h);
             if (result)
                 return result;
         }
@@ -197,10 +253,14 @@ EXPORT_SYM int chacha20_seek(stream_state *state,
                              unsigned offset)
 {
     int result;
+    uint32_t h[16];
 
     if (NULL == state)
         return ERR_NULL;
 
+    if ((state->nonceSize != 8) && (state->nonceSize != 12))
+        return ERR_NONCE_SIZE;
+
     if (offset >= sizeof state->keyStream)
         return ERR_MAX_OFFSET;
 
@@ -216,14 +276,48 @@ EXPORT_SYM int chacha20_seek(stream_state *state,
         state->h[12] = (uint32_t)block_low;
     }
 
-    result = chacha20_core(state);
+    result = chacha20_core(state, h);
     if (result)
         return result;
+
     state->usedKeyStream = offset;
 
     return 0;
 }
 
+/*
+ * Based on https://tools.ietf.org/html/draft-arciszewski-xchacha-03
+ */
+EXPORT_SYM int hchacha20(const uint8_t key[KEY_SIZE],
+                         const uint8_t nonce16[16],                 /* First 16 bytes of the 24 byte nonce */
+                         uint8_t subkey[KEY_SIZE])
+{
+    stream_state *pState;
+    uint32_t h[16];
+
+    if (NULL == key || NULL == nonce16 || NULL == subkey) {
+        return ERR_NULL;
+    }
+
+    chacha20_init(&pState, key, KEY_SIZE, nonce16, 16);
+    if (NULL == pState)
+        return ERR_MEMORY;
+
+    chacha20_core(pState, h);
+    /* We only keep first and last row from the new state */
+    STORE_U32_LITTLE(subkey + 0,  h[0]);
+    STORE_U32_LITTLE(subkey + 4,  h[1]);
+    STORE_U32_LITTLE(subkey + 8,  h[2]);
+    STORE_U32_LITTLE(subkey + 12, h[3]);
+    STORE_U32_LITTLE(subkey + 16, h[12]);
+    STORE_U32_LITTLE(subkey + 20, h[13]);
+    STORE_U32_LITTLE(subkey + 24, h[14]);
+    STORE_U32_LITTLE(subkey + 28, h[15]);
+    chacha20_destroy(pState);
+
+    return 0;
+}
+
 #ifdef PROFILE
 int main(void)
 {
@@ -242,7 +336,7 @@ int main(void)
 
     for (int i=0; i<1024; i++)
         chacha20_encrypt(state, data, data, data_size);
-    
+
     chacha20_destroy(state);
     free(data);
 }
